@@ -1,22 +1,22 @@
 /*
-   Copyright (C) 2003-2005 Samuel Mimram
-
-   This file is part of Ocaml-ssl.
-
-   This library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Lesser General Public
-   License as published by the Free Software Foundation; either
-   version 2.1 of the License, or (at your option) any later version.
-
-   This library is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with this library; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-*/
+ * Copyright (C) 2003-2005 Samuel Mimram
+ *
+ * This file is part of Ocaml-ssl.
+ *
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ *
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ */
 
 /**
  * Libssl bindings for OCaml.
@@ -24,23 +24,25 @@
  * @author Samuel Mimram
  */
 
-/* $Id: ssl_stubs.c,v 1.25 2005/07/18 23:34:18 smimram Exp $ */
+/* $Id: ssl_stubs.c 2684 2006-08-29 12:46:05Z smimram $ */
 
 /*
- * WARNING: because of thread callbacks, all ssl functions should be in blocking sections.
+ * WARNING: because of thread callbacks, all ssl functions should be in
+ * blocking sections.
  */
 
 #include <unistd.h>
-#include <strings.h>
+#include <string.h>
+#include <assert.h>
 
 #include <caml/alloc.h>
 #include <caml/callback.h>
+#include <caml/custom.h>
 #include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/mlvalues.h>
+#include <caml/signals.h>
 
-#include <string.h>
-#include <unistd.h>
 #include <openssl/ssl.h>
 #include <openssl/pem.h>
 #include <openssl/err.h>
@@ -52,36 +54,50 @@ static int client_verify_callback(int, X509_STORE_CTX *);
  * Data structures *
  *******************/
 
-struct ssl_socket__t
-{
-  SSL *handler;
-  int fd;
-};
+/* Contexts */
 
-typedef struct ssl_socket__t ssl_socket_t;
-
-static ssl_socket_t* ssl_socket_of_block(value block)
-{
-  return (ssl_socket_t*)Field(block, 1);
-}
-
-static void finalize_ssl_socket(value block)
-{
-  ssl_socket_t *s = ssl_socket_of_block(block);
-  caml_enter_blocking_section();
-  SSL_free(s->handler);
-  caml_leave_blocking_section();
-  free(s);
-}
+#define Ctx_val(v) (*((SSL_CTX**)Data_custom_val(v)))
 
 static void finalize_ctx(value block)
 {
-  SSL_CTX *ctx = (SSL_CTX*)Field(block, 1);
+  SSL_CTX *ctx = Ctx_val(block);
   caml_enter_blocking_section();
   SSL_CTX_free(ctx);
   caml_leave_blocking_section();
 }
 
+static struct custom_operations ctx_ops =
+{
+  "ocaml_ssl_ctx",
+  finalize_ctx,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
+};
+
+/* Sockets */
+
+#define SSL_val(v) (*((SSL**)Data_custom_val(v)))
+
+static void finalize_ssl_socket(value block)
+{
+  SSL *ssl = SSL_val(block);
+  caml_enter_blocking_section();
+  SSL_free(ssl);
+  caml_leave_blocking_section();
+}
+
+static struct custom_operations socket_ops =
+{
+  "ocaml_ssl_socket",
+  finalize_ssl_socket,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
+};
+
 
 /******************
  * Initialization *
@@ -89,8 +105,7 @@ static void finalize_ctx(value block)
 
 CAMLprim value ocaml_ssl_crypto_num_locks(value unit)
 {
-  CAMLparam1(unit);
-  CAMLreturn(Val_int(CRYPTO_num_locks()));
+  return Val_int(CRYPTO_num_locks());
 }
 
 void locking_function(int mode, int n, const char *file, int line)
@@ -110,13 +125,11 @@ void locking_function(int mode, int n, const char *file, int line)
 
 unsigned long id_function()
 {
-  return(Int_val(caml_callback(Field(Field(*caml_named_value("caml_ssl_thread_id_function"), 0), 0), Val_unit)));
+  return Int_val(caml_callback(Field(Field(*caml_named_value("caml_ssl_thread_id_function"), 0), 0), Val_unit));
 }
 
 CAMLprim value ocaml_ssl_init(value use_threads)
 {
-  CAMLparam1(use_threads);
-
   SSL_library_init();
   SSL_load_error_strings();
 
@@ -126,7 +139,14 @@ CAMLprim value ocaml_ssl_init(value use_threads)
     CRYPTO_set_locking_callback(locking_function);
   }
 
-  CAMLreturn(Val_unit);
+  return Val_unit;
+}
+
+CAMLprim value ocaml_ssl_get_error_string(value unit)
+{
+  char buf[256];
+  ERR_error_string_n(ERR_get_error(), buf, sizeof(buf));
+  return caml_copy_string(buf);
 }
 
 
@@ -134,11 +154,11 @@ CAMLprim value ocaml_ssl_init(value use_threads)
  * Context-related functions *
  *****************************/
 
-static SSL_METHOD *get_method(value protocol, int type)
+static SSL_METHOD *get_method(int protocol, int type)
 {
-  SSL_METHOD *method;
+  SSL_METHOD *method = NULL;
 
-  switch (Int_val(protocol))
+  switch (protocol)
   {
     case 0:
       switch (type)
@@ -221,25 +241,29 @@ static SSL_METHOD *get_method(value protocol, int type)
 
 CAMLprim value ocaml_ssl_create_context(value protocol, value type)
 {
-  CAMLparam2(protocol, type);
+  value block;
   SSL_CTX *ctx;
-  CAMLlocal1(block);
+  SSL_METHOD *method = get_method(Int_val(protocol), Int_val(type));
 
   caml_enter_blocking_section();
-  ctx = SSL_CTX_new(get_method(protocol, Int_val(type)));
+  ctx = SSL_CTX_new(method);
   caml_leave_blocking_section();
   if (!ctx)
     caml_raise_constant(*caml_named_value("ssl_exn_context_error"));
+  /* In non-blocking mode, accept a buffer with a different address on
+     a write retry (since the GC may need to move it). In blocking
+     mode, hide SSL_ERROR_WANT_(READ|WRITE) from us. */
+  SSL_CTX_set_mode(ctx, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER | SSL_MODE_AUTO_RETRY);
 
-  block = caml_alloc_final(2, finalize_ctx, 0, 1);
-  Store_field(block, 1, (value)ctx);
-  CAMLreturn(block);
+  block = caml_alloc_custom(&ctx_ops, sizeof(SSL_CTX*), 0, 1);
+  Ctx_val(block) = ctx;
+  return block;
 }
 
-CAMLprim ocaml_ssl_ctx_use_certificate(value context, value cert, value privkey)
+CAMLprim value ocaml_ssl_ctx_use_certificate(value context, value cert, value privkey)
 {
   CAMLparam3(context, cert, privkey);
-  SSL_CTX *ctx = (SSL_CTX*)Field(context, 1);
+  SSL_CTX *ctx = Ctx_val(context);
   char *cert_name = String_val(cert);
   char *privkey_name = String_val(privkey);
 
@@ -264,13 +288,14 @@ CAMLprim ocaml_ssl_ctx_use_certificate(value context, value cert, value privkey)
   CAMLreturn(Val_unit);
 }
 
-CAMLprim ocaml_ssl_get_verify_result(value socket)
+CAMLprim value ocaml_ssl_get_verify_result(value socket)
 {
   CAMLparam1(socket);
   int ans;
+  SSL *ssl = SSL_val(socket);
 
   caml_enter_blocking_section();
-  ans = SSL_get_verify_result(ssl_socket_of_block(socket)->handler);
+  ans = SSL_get_verify_result(ssl);
   caml_leave_blocking_section();
 
   CAMLreturn(Val_int(ans));
@@ -278,15 +303,13 @@ CAMLprim ocaml_ssl_get_verify_result(value socket)
 
 CAMLprim value ocaml_ssl_get_client_verify_callback_ptr(value unit)
 {
-  CAMLparam1(unit);
-
-  CAMLreturn((value)client_verify_callback);
+  return (value)client_verify_callback;
 }
 
 CAMLprim value ocaml_ssl_ctx_set_verify(value context, value vmode, value vcallback)
 {
   CAMLparam3(context, vmode, vcallback);
-  SSL_CTX *ctx = (SSL_CTX*)Field(context, 1);
+  SSL_CTX *ctx = Ctx_val(context);
   int mode = 0;
   value mode_tl = vmode;
   int (*callback) (int, X509_STORE_CTX*) = NULL;
@@ -329,8 +352,7 @@ CAMLprim value ocaml_ssl_ctx_set_verify(value context, value vmode, value vcallb
 
 CAMLprim value ocaml_ssl_ctx_set_verify_depth(value context, value vdepth)
 {
-  CAMLparam2(context, vdepth);
-  SSL_CTX *ctx = (SSL_CTX*)Field(context, 1);
+  SSL_CTX *ctx = Ctx_val(context);
   int depth = Int_val(vdepth);
 
   if (depth < 0)
@@ -340,13 +362,13 @@ CAMLprim value ocaml_ssl_ctx_set_verify_depth(value context, value vdepth)
   SSL_CTX_set_verify_depth(ctx, depth);
   caml_leave_blocking_section();
 
-  CAMLreturn(Val_unit);
+  return Val_unit;
 }
 
 CAMLprim value ocaml_ssl_ctx_set_client_CA_list_from_file(value context, value vfilename)
 {
   CAMLparam2(context, vfilename);
-  SSL_CTX *ctx = (SSL_CTX*)Field(context, 1);
+  SSL_CTX *ctx = Ctx_val(context);
   char *filename = String_val(vfilename);
   STACK_OF(X509_NAME) *cert_names;
 
@@ -364,15 +386,48 @@ CAMLprim value ocaml_ssl_ctx_set_client_CA_list_from_file(value context, value v
   CAMLreturn(Val_unit);
 }
 
+static int pem_passwd_cb(char *buf, int size, int rwflag, void *userdata)
+{
+  value s;
+  int len;
+
+  caml_leave_blocking_section();
+  s = caml_callback(*((value*)userdata), Val_int(rwflag));
+  len = caml_string_length(s);
+  assert(len <= size);
+  memcpy(buf, String_val(s), len);
+  caml_enter_blocking_section();
+
+  return len;
+}
+
+CAMLprim value ocaml_ssl_ctx_set_default_passwd_cb(value context, value cb)
+{
+  CAMLparam2(context, cb);
+  SSL_CTX *ctx = Ctx_val(context);
+  value *pcb;
+
+  /* TODO: this never gets freed or even unregistered */
+  pcb = malloc(sizeof(value));
+  *pcb = cb;
+  caml_register_global_root(pcb);
+
+  caml_enter_blocking_section();
+  SSL_CTX_set_default_passwd_cb(ctx, pem_passwd_cb);
+  SSL_CTX_set_default_passwd_cb_userdata(ctx, pcb);
+  caml_leave_blocking_section();
+
+  CAMLreturn(Val_unit);
+}
 
 /****************************
  * Cipher-related functions *
  ****************************/
 
-CAMLprim ocaml_ssl_ctx_set_cipher_list(value context, value ciphers_string)
+CAMLprim value ocaml_ssl_ctx_set_cipher_list(value context, value ciphers_string)
 {
   CAMLparam2(context, ciphers_string);
-  SSL_CTX *ctx = (SSL_CTX*)Field(context, 1);
+  SSL_CTX *ctx = Ctx_val(context);
   char *ciphers = String_val(ciphers_string);
 
   if(*ciphers == 0)
@@ -392,10 +447,10 @@ CAMLprim ocaml_ssl_ctx_set_cipher_list(value context, value ciphers_string)
 CAMLprim value ocaml_ssl_get_current_cipher(value socket)
 {
   CAMLparam1(socket);
-  SSL *handler = ssl_socket_of_block(socket)->handler;
+  SSL *ssl = SSL_val(socket);
 
   caml_enter_blocking_section();
-  SSL_CIPHER *cipher = (SSL_CIPHER*)SSL_get_current_cipher(handler);
+  SSL_CIPHER *cipher = (SSL_CIPHER*)SSL_get_current_cipher(ssl);
   caml_leave_blocking_section();
   if (!cipher)
     caml_raise_constant(*caml_named_value("ssl_exn_cipher_error"));
@@ -405,7 +460,6 @@ CAMLprim value ocaml_ssl_get_current_cipher(value socket)
 
 CAMLprim value ocaml_ssl_get_cipher_description(value vcipher)
 {
-  CAMLparam1(vcipher);
   char buf[1024];
   SSL_CIPHER *cipher = (SSL_CIPHER*)vcipher;
 
@@ -413,12 +467,11 @@ CAMLprim value ocaml_ssl_get_cipher_description(value vcipher)
   SSL_CIPHER_description(cipher, buf, 1024);
   caml_leave_blocking_section();
 
-  CAMLreturn(caml_copy_string(buf));
+  return caml_copy_string(buf);
 }
 
 CAMLprim value ocaml_ssl_get_cipher_name(value vcipher)
 {
-  CAMLparam1(vcipher);
   const char *name;
   SSL_CIPHER *cipher = (SSL_CIPHER*)vcipher;
 
@@ -426,12 +479,11 @@ CAMLprim value ocaml_ssl_get_cipher_name(value vcipher)
   name = SSL_CIPHER_get_name(cipher);
   caml_leave_blocking_section();
 
-  CAMLreturn(caml_copy_string(name));
+  return caml_copy_string(name);
 }
 
 CAMLprim value ocaml_ssl_get_cipher_version(value vcipher)
 {
-  CAMLparam1(vcipher);
   char *version;
   SSL_CIPHER *cipher = (SSL_CIPHER*)vcipher;
 
@@ -439,24 +491,36 @@ CAMLprim value ocaml_ssl_get_cipher_version(value vcipher)
   version = SSL_CIPHER_get_version(cipher);
   caml_leave_blocking_section();
 
-  CAMLreturn(caml_copy_string(version));
+  return caml_copy_string(version);
 }
 
 /*********************************
  * Certificate-related functions *
  *********************************/
 
+#define Cert_val(v) (*((X509**)Data_custom_val(v)))
+
 static void finalize_cert(value block)
 {
-  X509 *cert = (X509*)Field(block, 1);
+  X509 *cert = Cert_val(block);
   caml_enter_blocking_section();
   X509_free(cert);
   caml_leave_blocking_section();
 }
 
-CAMLprim ocaml_ssl_read_certificate(value vfilename)
+static struct custom_operations cert_ops =
 {
-  CAMLparam1(vfilename);
+  "ocaml_ssl_cert",
+  finalize_cert,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
+};
+
+CAMLprim value ocaml_ssl_read_certificate(value vfilename)
+{
+  value block;
   char *filename = String_val(vfilename);
   X509 *cert = NULL;
   FILE *fh = NULL;
@@ -471,21 +535,21 @@ CAMLprim ocaml_ssl_read_certificate(value vfilename)
     caml_leave_blocking_section();
     caml_raise_constant(*caml_named_value("ssl_exn_certificate_error"));
   }
+  fclose(fh);
   caml_leave_blocking_section();
 
-  CAMLlocal1(block);
-  block = caml_alloc_final(2, finalize_cert, 0, 1);
-  Store_field(block, 1, (value)cert);
-  CAMLreturn(block);
+  block = caml_alloc_custom(&cert_ops, sizeof(X509*), 0, 1);
+  Cert_val(block) = cert;
+  return block;
 }
 
 CAMLprim value ocaml_ssl_get_certificate(value socket)
 {
   CAMLparam1(socket);
-  SSL *handler = ssl_socket_of_block(socket)->handler;
+  SSL *ssl = SSL_val(socket);
 
   caml_enter_blocking_section();
-  X509 *cert = SSL_get_peer_certificate(handler);
+  X509 *cert = SSL_get_peer_certificate(ssl);
   caml_leave_blocking_section();
 
   if (!cert)
@@ -500,7 +564,7 @@ CAMLprim value ocaml_ssl_get_certificate(value socket)
 CAMLprim value ocaml_ssl_get_issuer(value certificate)
 {
   CAMLparam1(certificate);
-  X509 *cert = (X509*)Field(certificate, 1);
+  X509 *cert = Cert_val(certificate);
 
   caml_enter_blocking_section();
   char *issuer = X509_NAME_oneline(X509_get_issuer_name(cert), 0, 0);
@@ -513,7 +577,7 @@ CAMLprim value ocaml_ssl_get_issuer(value certificate)
 CAMLprim value ocaml_ssl_get_subject(value certificate)
 {
   CAMLparam1(certificate);
-  X509 *cert = (X509*)Field(certificate, 1);
+  X509 *cert = Cert_val(certificate);
 
   caml_enter_blocking_section();
   char *subject = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0);
@@ -523,10 +587,10 @@ CAMLprim value ocaml_ssl_get_subject(value certificate)
   CAMLreturn(caml_copy_string(subject));
 }
 
-CAMLprim ocaml_ssl_ctx_load_verify_locations(value context, value ca_file, value ca_path)
+CAMLprim value ocaml_ssl_ctx_load_verify_locations(value context, value ca_file, value ca_path)
 {
   CAMLparam3(context, ca_file, ca_path);
-  SSL_CTX *ctx = (SSL_CTX*)Field(context, 1);
+  SSL_CTX *ctx = Ctx_val(context);
   char *CAfile = String_val(ca_file);
   char *CApath = String_val(ca_path);
 
@@ -553,34 +617,38 @@ CAMLprim ocaml_ssl_ctx_load_verify_locations(value context, value ca_file, value
 CAMLprim value ocaml_ssl_get_file_descr(value socket)
 {
   CAMLparam1(socket);
+  SSL *ssl = SSL_val(socket);
+  int fd;
 
-  CAMLreturn(Val_int(ssl_socket_of_block(socket)->fd));
+  caml_enter_blocking_section();
+  fd = SSL_get_fd(ssl);
+  caml_leave_blocking_section();
+
+  CAMLreturn(Val_int(fd));
 }
 
 CAMLprim value ocaml_ssl_embed_socket(value socket_, value context)
 {
-  CAMLparam2(socket_, context);
-  int socket = Int_val(socket_);
-  SSL_CTX *ctx = (SSL_CTX*)Field(context, 1);
-
-  ssl_socket_t *ret = malloc(sizeof(ssl_socket_t));
+  CAMLparam1(context);
   CAMLlocal1(block);
+  int socket = Int_val(socket_);
+  SSL_CTX *ctx = Ctx_val(context);
+  SSL *ssl;
+
+  block = caml_alloc_custom(&socket_ops, sizeof(SSL*), 0, 1);
 
   if (socket < 0)
     caml_raise_constant(*caml_named_value("ssl_exn_invalid_socket"));
   caml_enter_blocking_section();
-  ret->handler = SSL_new(ctx);
-  if (!ret->handler)
+  ssl = SSL_new(ctx);
+  if (!ssl)
   {
     caml_leave_blocking_section();
     caml_raise_constant(*caml_named_value("ssl_exn_handler_error"));
   }
-  SSL_set_fd(ret->handler, socket);
+  SSL_set_fd(ssl, socket);
   caml_leave_blocking_section();
-  ret->fd = socket;
-
-  block = caml_alloc_final(2, finalize_ssl_socket, 0, 1);
-  Store_field(block, 1, (value)ret);
+  SSL_val(block) = ssl;
 
   CAMLreturn(block);
 }
@@ -589,17 +657,17 @@ CAMLprim value ocaml_ssl_connect(value socket)
 {
   CAMLparam1(socket);
   int ret;
-  SSL *handler = ssl_socket_of_block(socket)->handler;
+  SSL *ssl = SSL_val(socket);
 
   caml_enter_blocking_section();
-  ret = SSL_connect(handler);
+  ret = SSL_connect(ssl);
   caml_leave_blocking_section();
-  if (ret == -1)
+  if (ret < 0)
   {
     int err;
 
     caml_enter_blocking_section();
-    err = SSL_get_error(ssl_socket_of_block(socket)->handler, ret);
+    err = SSL_get_error(ssl, ret);
     caml_leave_blocking_section();
     caml_raise_with_arg(*caml_named_value("ssl_exn_connection_error"), Val_int(err));
   }
@@ -607,13 +675,14 @@ CAMLprim value ocaml_ssl_connect(value socket)
   CAMLreturn(Val_unit);
 }
 
-CAMLprim ocaml_ssl_verify(value socket)
+CAMLprim value ocaml_ssl_verify(value socket)
 {
   CAMLparam1(socket);
+  SSL *ssl = SSL_val(socket);
   long ans;
 
   caml_enter_blocking_section();
-  ans = SSL_get_verify_result(ssl_socket_of_block(socket)->handler);
+  ans = SSL_get_verify_result(ssl);
   caml_leave_blocking_section();
 
   if (ans != 0)
@@ -629,19 +698,19 @@ CAMLprim ocaml_ssl_verify(value socket)
 
 CAMLprim value ocaml_ssl_write(value socket, value buffer, value start, value length)
 {
-  CAMLparam4(socket, buffer, start, length);
+  CAMLparam2(socket, buffer);
   int ret, err;
   int buflen = Int_val(length);
   char *buf = malloc(buflen);
-  SSL *handler = ssl_socket_of_block(socket)->handler;
+  SSL *ssl = SSL_val(socket);
 
-  if (caml_string_length(buffer) - Int_val(start) < Int_val(length))
+  if (Int_val(start) + Int_val(length) > caml_string_length(buffer))
     caml_invalid_argument("Buffer too short.");
 
   memmove(buf, (char*)String_val(buffer) + Int_val(start), buflen);
   caml_enter_blocking_section();
-  ret = SSL_write(handler, buf, buflen);
-  err = SSL_get_error(handler, ret);
+  ret = SSL_write(ssl, buf, buflen);
+  err = SSL_get_error(ssl, ret);
   caml_leave_blocking_section();
   free(buf);
 
@@ -653,16 +722,20 @@ CAMLprim value ocaml_ssl_write(value socket, value buffer, value start, value le
 
 CAMLprim value ocaml_ssl_read(value socket, value buffer, value start, value length)
 {
-  CAMLparam4(socket, buffer, start, length);
+  CAMLparam2(socket, buffer);
   int ret, err;
   int buflen = Int_val(length);
   char *buf = malloc(buflen);
-  SSL *handler = ssl_socket_of_block(socket)->handler;
+  SSL *ssl = SSL_val(socket);
+
+  if (Int_val(start) + Int_val(length) > caml_string_length(buffer))
+    caml_invalid_argument("Buffer too short.");
 
   caml_enter_blocking_section();
-  ret = SSL_read(handler, buf, buflen);
+  ret = SSL_read(ssl, buf, buflen);
+  err = SSL_get_error(ssl, ret);
   if (err != SSL_ERROR_NONE)
-    err = SSL_get_error(handler, ret);
+    err = SSL_get_error(ssl, ret);
   caml_leave_blocking_section();
   memmove(((char*)String_val(buffer)) + Int_val(start), buf, buflen);
   free(buf);
@@ -676,14 +749,14 @@ CAMLprim value ocaml_ssl_read(value socket, value buffer, value start, value len
 CAMLprim value ocaml_ssl_accept(value socket)
 {
   CAMLparam1(socket);
-  SSL *handler = ssl_socket_of_block(socket)->handler;
+  SSL *ssl = SSL_val(socket);
 
   int ret, err;
   caml_enter_blocking_section();
-  ret = SSL_accept(handler);
+  ret = SSL_accept(ssl);
   if (ret <= 0)
   {
-    err = SSL_get_error(handler, ret);
+    err = SSL_get_error(ssl, ret);
     caml_leave_blocking_section();
     caml_raise_with_arg(*caml_named_value("ssl_exn_accept_error"), Val_int(err));
   }
@@ -695,15 +768,17 @@ CAMLprim value ocaml_ssl_accept(value socket)
 CAMLprim value ocaml_ssl_flush(value socket)
 {
   CAMLparam1(socket);
-  ssl_socket_t *ssl = ssl_socket_of_block(socket);
+  SSL *ssl = SSL_val(socket);
+  BIO *bio;
+  int ans;
 
-  BIO *bio = SSL_get_wbio(ssl_socket_of_block(socket)->handler);
+  caml_enter_blocking_section();
+  bio = SSL_get_wbio(ssl);
   if(bio)
   {
-    caml_enter_blocking_section();
-    BIO_flush(bio);
-    caml_leave_blocking_section();
+    ans = BIO_flush(bio);
   }
+  caml_leave_blocking_section();
 
   CAMLreturn(Val_unit);
 }
@@ -711,15 +786,15 @@ CAMLprim value ocaml_ssl_flush(value socket)
 CAMLprim value ocaml_ssl_shutdown(value socket)
 {
   CAMLparam1(socket);
-  ssl_socket_t *ssl = ssl_socket_of_block(socket);
+  SSL *ssl = SSL_val(socket);
   int ret;
 
   caml_enter_blocking_section();
-  ret = SSL_shutdown(ssl->handler);
+  ret = SSL_shutdown(ssl);
   if (!ret)
-    SSL_shutdown(ssl->handler);
+    SSL_shutdown(ssl);
   caml_leave_blocking_section();
-  close(ssl->fd);
+  /* close(SSL_get_fd(SSL_val(socket))); */
 
   CAMLreturn(Val_unit);
 }
@@ -768,7 +843,6 @@ CAMLprim value ocaml_ssl_shutdown(value socket)
 
 static int client_verify_callback(int ok, X509_STORE_CTX *ctx)
 {
-  X509 *peer;
   char *subject, *issuer;
   int depth, error;
   char *xs;
